import textwrap

import karathon
from karabo.common.scenemodel.api import (
    CheckBoxModel,
    ColorBoolModel,
    DeviceSceneLinkModel,
    DisplayCommandModel,
    DisplayLabelModel,
    DisplayStateColorModel,
    DisplayTextLogModel,
    DoubleLineEditModel,
    EvaluatorModel,
    IntLineEditModel,
    LineEditModel,
    LabelModel,
    SceneModel,
    write_scene,
    RectangleModel,
)

BASE_INC = 30
PADDING = 10
RECONFIGURABLE = 4  # TODO: look up proper enum

_type_to_line_editable = {
    "DOUBLE": (DoubleLineEditModel, {}),
    "FLOAT": (DoubleLineEditModel, {}),
    "INT32": (IntLineEditModel, {}),
    "UINT32": (IntLineEditModel, {}),
    "INT64": (IntLineEditModel, {}),
    "UINT64": (IntLineEditModel, {}),
    "STRING": (LineEditModel, {"klass": "EditableLineEdit"}),
}


def titled(title, width=8):
    def actual_decorator(component_class):
        # should this create subclass instead of mutating?
        orig_render = component_class.render

        def new_render(self, x, y, *args, **kwargs):
            return [
                LabelModel(
                    frame_width=1,
                    text=title,
                    width=width * BASE_INC,
                    height=BASE_INC,
                    x=x,
                    y=y,
                )
            ] + orig_render(self, x, y + BASE_INC, *args, **kwargs)

        component_class.render = new_render

        orig_height = component_class.height

        def new_height(self):
            return orig_height(self) + BASE_INC

        component_class.height = new_height

        return component_class

    return actual_decorator


def boxed(component_class):
    # should this create subclass instead of mutating?
    orig_render = component_class.render
    orig_height = component_class.height
    orig_width = component_class.width

    def new_render(self, x, y, *args, **kwargs):
        return [
            RectangleModel(
                x=x,
                y=y,
                width=orig_width(self) + 2 * PADDING,
                height=orig_height(self) + 2 * PADDING,
                stroke="#000000",
            )
        ] + orig_render(self, x + PADDING, y + PADDING, *args, **kwargs)

    component_class.render = new_render

    def new_height(self):
        return orig_height(self) + 2 * PADDING

    component_class.height = new_height

    def new_width(self):
        return orig_width(self) + 2 * PADDING

    component_class.width = new_width

    return component_class


class MaybeEditableRow:
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7,
        display_width=5,
        edit_width=5,
    ):
        self.label_width = label_width * BASE_INC
        self.display_width = display_width * BASE_INC
        self.edit_width = edit_width * BASE_INC

        key_attr = schema_hash.getAttributes(key_path)
        label_text = textwrap.shorten(
            key_attr["displayedName"]
            if "displayedName" in key_attr
            else key_path.split(".")[-1],
            24,
        )

        self.label = LabelModel(
            text=label_text,
            width=self.label_width,
            height=BASE_INC,
        )
        self.value_display = DisplayLabelModel(
            keys=[f"{device_id}.{key_path}"],
            width=self.display_width,
            height=BASE_INC,
        )
        if key_attr["accessMode"] == RECONFIGURABLE:
            value_type = key_attr["valueType"]
            if value_type in _type_to_line_editable:
                line_editable_class, extra_args = _type_to_line_editable[value_type]
                self.value_edit = line_editable_class(
                    keys=[f"{device_id}.{key_path}"],
                    width=self.edit_width,
                    height=BASE_INC,
                    **extra_args,
                )
            else:
                self.value_edit = LabelModel(
                    text=f"Not yet supported: editing {key_path} of type {value_type}",
                    width=self.edit_width,
                    height=BASE_INC,
                )

    def height(self):
        return BASE_INC

    def width(self):
        return self.label_width + self.display_width + self.edit_width

    def render(self, x=None, y=None):
        self.label.x = x
        self.label.y = y
        self.value_display.x = x + self.label_width
        self.value_display.y = y
        if hasattr(self, "value_edit"):
            self.value_edit.x = x + self.label_width + self.display_width
            self.value_edit.y = y
            return [
                self.label,
                self.value_display,
                self.value_edit,
            ]
        else:
            return [
                self.label,
                self.value_display,
            ]


@titled("Parameters used for CalCat queries", width=10)
@boxed
class ConstantParameterColumn:
    def __init__(self, device_id, schema_hash, prefix="constantParameters"):
        self.device_id = device_id
        self.rows = [
            MaybeEditableRow(device_id, schema_hash, f"{prefix}.{key}")
            for key in schema_hash.get(prefix).getKeys()
        ]

    def render(self, x, y):
        res = []
        y_offset = 0
        for row in self.rows:
            res.extend(row.render(x=x, y=y + y_offset))
            y_offset += row.height()

        res.append(
            DisplayCommandModel(
                keys=[f"{self.device_id}.loadMostRecentConstants"],
                x=x + 7 * BASE_INC,
                y=y + y_offset,
                width=10 * BASE_INC,
                height=BASE_INC,
            )
        )

        return res

    def height(self):
        return sum(row.height() for row in self.rows) + BASE_INC

    def width(self):
        return max(row.width() for row in self.rows)


class ConstantNode:
    def __init__(self, device_id, constant_path):
        self.title = LabelModel(
            text=constant_path.split(".")[-1],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.ampel = ColorBoolModel(
            height=BASE_INC, width=BASE_INC, keys=[f"{device_id}.{constant_path}.found"]
        )

    def render(self, x, y):
        self.title.x = x
        self.title.y = y
        self.ampel.x = x + 7 * BASE_INC
        self.ampel.y = y
        y += BASE_INC
        return [self.title, self.ampel]

    def height(self):
        return BASE_INC

    def width(self):
        return 8 * BASE_INC


@titled("Found constants", width=6)
@boxed
class FoundConstantsColumn:
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        self.rows = [
            ConstantNode(device_id, f"{prefix}.{constant_name}")
            for constant_name in schema_hash.get(prefix).getKeys()
        ]

    def render(self, x, y):
        res = []
        y_offset = 0
        for row in self.rows:
            res.extend(row.render(x, y + y_offset))
            y_offset += row.height()
        return res

    def height(self):
        return sum(row.height() for row in self.rows)

    def width(self):
        return max(row.width() for row in self.rows)


class CorrectionStepNode:
    def __init__(self, device_id, step_path):
        self.label = LabelModel(
            text=step_path.split(".")[-1], width=7 * BASE_INC, height=BASE_INC
        )
        self.checkbox_main = CheckBoxModel(
            keys=[f"{device_id}.{step_path}.enable"],
            height=BASE_INC,
            width=BASE_INC,
            klass="EditableCheckBox",
        )
        self.checkbox_preview = CheckBoxModel(
            keys=[f"{device_id}.{step_path}.preview"],
            height=BASE_INC,
            width=BASE_INC,
            klass="EditableCheckBox",
        )

    def render(self, x, y):
        self.label.x = x
        self.label.y = y
        x += 5 * BASE_INC
        self.checkbox_main.x = x
        self.checkbox_main.y = y
        x += 3 * BASE_INC
        self.checkbox_preview.x = x
        self.checkbox_preview.y = y
        return [self.label, self.checkbox_main, self.checkbox_preview]

    def height(self):
        return BASE_INC

    def width(self):
        return 10 * BASE_INC


@titled("Correction steps", width=6)
@boxed
class CorrectionStepsColumn:
    def __init__(
        self, device_id, schema_hash, prefix="corrections", control_prefix=None
    ):
        # separate control prefix allows iterating device schema, but setting on manager
        if control_prefix is None:
            control_prefix = prefix

        self.header_main = LabelModel(
            text="enable", width=3 * BASE_INC, height=BASE_INC
        )
        self.header_preview = LabelModel(
            text="preview", width=3 * BASE_INC, height=BASE_INC
        )
        self.steps = [
            CorrectionStepNode(device_id, f"{control_prefix}.{step_path}")
            for step_path in schema_hash.get(prefix).getKeys()
        ]

    def render(self, x, y):
        self.header_main.x = x + 5 * BASE_INC
        self.header_main.y = y
        self.header_preview.x = x + 8 * BASE_INC
        self.header_preview.y = y
        res = [self.header_main, self.header_preview]
        y += BASE_INC
        for step in self.steps:
            step.x = x
            step.y = y
            res.extend(step.render(x, y))
            y += step.height()
        return res

    def width(self):
        return max(step.width() for step in self.steps)

    def height(self):
        return sum(step.height() for step in self.steps) + BASE_INC


class ConstantLoadedAmpeln:
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        self.keys = [
            f"{device_id}.{prefix}.{key}.found"
            for key in schema_hash.get(prefix).getKeys()
        ]

    def render(self, x, y):
        return [
            ColorBoolModel(
                x=x + i * BASE_INC,
                y=y,
                height=BASE_INC,
                width=BASE_INC,
                keys=[key],
            )
            for i, key in enumerate(self.keys)
        ]

    def width(self):
        return BASE_INC * len(self.keys)

    def height(self):
        return BASE_INC


@titled("Device status", width=6)
@boxed
class CorrectionDeviceStatus:
    def __init__(self, device_id):
        self.name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        self.state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.rate = EvaluatorModel(
            expression="f'{x:.02f}'",
            keys=[f"{device_id}.performance.rate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.processing_time = EvaluatorModel(
            expression="f'{x:.02f}'",
            keys=[f"{device_id}.performance.processingDuration"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=20 * BASE_INC,
        )

    def render(self, x, y):
        self.name.x = x
        self.name.y = y
        y += BASE_INC
        self.state.x = x
        self.state.y = y
        self.tid.x = x + 7 * BASE_INC
        self.tid.y = y
        y += BASE_INC
        self.rate.x = x
        self.rate.y = y
        self.processing_time.x = x + 7 * BASE_INC
        self.processing_time.y = y
        y += BASE_INC
        self.status_log.x = x
        self.status_log.y = y
        return [
            self.name,
            self.state,
            self.rate,
            self.processing_time,
            self.tid,
            self.status_log,
        ]

    def width(self):
        return 14 * BASE_INC

    def height(self):
        return 23 * BASE_INC


class CompactCorrectionDeviceOverview:
    def __init__(self, device_id, schema_hash):
        self.name = DeviceSceneLinkModel(
            text=device_id.split("/")[-1],
            keys=[f"{device_id}.availableScenes"],
            target="overview",
            width=5 * BASE_INC,
            height=BASE_INC
        )
        self.status = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=6 * BASE_INC,
            height=BASE_INC,
        )
        self.rate = EvaluatorModel(
            expression="f'{x:.02f}'",
            keys=[f"{device_id}.performance.rate"],
            width=4 * BASE_INC,
            height=BASE_INC,
        )
        self.tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=4 * BASE_INC,
            height=BASE_INC,
        )
        self.ampeln = ConstantLoadedAmpeln(device_id, schema_hash)

    def render(self, x, y):
        self.name.x = x
        self.name.y = y
        x += self.name.width
        self.status.x = x
        self.status.y = y
        x += self.status.width
        self.rate.x = x
        self.rate.y = y
        x += self.rate.width
        self.tid.x = x
        self.tid.y = y
        x += self.tid.width
        return [self.name, self.status, self.rate, self.tid] + self.ampeln.render(x, y)

    def width(self):
        return 19 * BASE_INC + self.ampeln.width()

    def height(self):
        return BASE_INC


def correction_device_overview_scene(device_id, schema):
    if isinstance(schema, karathon.Schema):
        schema_hash = schema.getParameterHash()
    else:
        schema_hash = schema.hash

    status_overview = CorrectionDeviceStatus(device_id)
    cpc = ConstantParameterColumn(device_id, schema_hash)
    fcc = FoundConstantsColumn(device_id, schema_hash)
    csc = CorrectionStepsColumn(device_id, schema_hash)

    subscenes = []
    x = PADDING
    y = PADDING
    subscenes.extend(status_overview.render(x, y))
    x += status_overview.width() + PADDING
    subscenes.extend(cpc.render(x, y))
    x += cpc.width() + PADDING
    subscenes.extend(fcc.render(x, y))
    y += fcc.height() + PADDING
    subscenes.extend(csc.render(x, y))

    scene = SceneModel(
        height=max(
            status_overview.height(),
            cpc.height(),
            fcc.height() + PADDING + csc.height(),
        )
        + 2 * PADDING,
        width=2 * PADDING
        + 2 * PADDING
        + status_overview.width()
        + cpc.width()
        + csc.width(),
        children=subscenes,
    )
    return write_scene(scene)


def manager_device_overview_scene(
    manager_device_id, correction_device_schema, correction_device_ids
):
    if isinstance(correction_device_schema, karathon.Schema):
        schema_hash = correction_device_schema.getParameterHash()
    else:
        schema_hash = correction_device_schema.hash

    x = PADDING
    y = PADDING
    subscenes = []
    correction_steps = CorrectionStepsColumn(
        manager_device_id, schema_hash, control_prefix="managed.corrections"
    )
    subscenes.extend(correction_steps.render(x, y))
    y += correction_steps.height() + PADDING
    for device_id in correction_device_ids:
        ccdo = CompactCorrectionDeviceOverview(device_id, schema_hash)
        subscenes.extend(ccdo.render(x, y))
        y += ccdo.height()

    scene = SceneModel(
        height=y + PADDING,
        width=ccdo.width() + 2 * PADDING,
        children=subscenes,
    )
    return write_scene(scene)
